//#include "stdafx.h"
#include "../cmath3d_v/TriangleMesh_v.h"
#include <map>

namespace ulm
{

//----------- vertex normal calculation functions -----------

void getVertexNormalviaBarycentre(const ActiveMesh& inMesh,
                                  const size_t inV,
                                  Vector3F& norm,
                                  const size_t maxLevel)
{
	//skip on incorrect input
	if (inV >= (signed)inMesh.Pos.GetLen()) return;

	//entry point to the list of vertices
	const Vector3FC* const Pos=inMesh.Pos.GetPtr();

	//list of surrounding neighbours to include in the calculation
	std::vector< std::vector<size_t> > neigs;
	getVertexNeighbours(inMesh,inV,maxLevel,neigs);

	//nullify the output norm
	norm.Reset();

	//process surrounding neighbours with various weight
	float curWeight=1.0f;
	Vector3F localNorm;

	//over surrounding levels...
	for (size_t i=0; i < neigs.size(); ++i)
	{
		//over all vertices at this level
		for (size_t j=0; j < neigs[i].size(); ++j)
		{
			std::vector<size_t> n;
			getVertexImmediateNeighbours(inMesh,neigs[i][j],n);

			const Vector3FC& v=Pos[neigs[i][j]];

			//over all neighbours of the current vertex
			localNorm.Reset();
			for (size_t k=0; k < n.size(); ++k)
			{
				Vector3FC nV=Pos[n[k]];
				nV-=v;
				localNorm+=nV;
			}
			localNorm/=(n.size() > 0)? float(n.size()) : 1.f;

			localNorm/=curWeight * float(neigs[i].size());
			norm+=localNorm;
		}

		curWeight*=(float)(i+2);
	}

	//norm*=-1.f;
}


void getVertexNormalviaTNormals(const ActiveMesh& inMesh,
                                const size_t inV,
                                Vector3F& norm,
                                const int mode)
{
	//skip on incorrect input
	if (inV >= (signed)inMesh.Pos.GetLen()) return;

	//entry points to the lists of vertices and triangles
	const Vector3FC* const Pos=inMesh.Pos.GetPtr();
	const DWORD* const ID=inMesh.ID.GetPtr();

	//list of surrounding triangles to include in the calculation
	std::vector<size_t> neigsT;
	getVertexIncidentTriangles(inMesh,inV,neigsT);

	//helper variables...
	float w=0;
	Vector3F tmp;

	//nullify the output norm
	norm.Reset();

	if (mode == AreaWeight)
	{
		for (size_t i=0; i < neigsT.size(); ++i)
		{
			//determine triangle normal and size
			const size_t v0=ID[neigsT[i]*3+0];
			const size_t v1=ID[neigsT[i]*3+1];
			const size_t v2=ID[neigsT[i]*3+2];

			//MT template for calculating normals: 2-1 x 0-2
			Mul(Pos[v2]-Pos[v1],Pos[v0]-Pos[v2],tmp);

			float area=tmp.Len();
			norm+=tmp;
			w+=area;
		}
	}
	else
	if (mode == AngleWeight)
	{
		for (size_t i=0; i < neigsT.size(); ++i)
		{
			//determine triangle normal and size
			const size_t v0=ID[neigsT[i]*3+0];
			const size_t v1=ID[neigsT[i]*3+1];
			const size_t v2=ID[neigsT[i]*3+2];

			float angle;
			if (inV == v0)
			{
				tmp=Pos[v1]-Pos[v0];
				angle=tmp.Angle(Pos[v2]-Pos[v0]);
			}
			else
			if (inV == v1)
			{
				tmp=Pos[v0]-Pos[v1];
				angle=tmp.Angle(Pos[v2]-Pos[v1]);
			}
			else
			//inV == v2
			{
				tmp=Pos[v0]-Pos[v2];
				angle=tmp.Angle(Pos[v1]-Pos[v2]);
			}

			//MT template for calculating normals: 2-1 x 0-2
			Mul(Pos[v2]-Pos[v1],Pos[v0]-Pos[v2],tmp);

			norm+=angle*tmp.Ort();
			w+=angle;
		}
		//note: 'w' should be more-or-less equal to 2PI, right?
	}
	else //no weights mode
	{
		for (size_t i=0; i < neigsT.size(); ++i)
		{
			//determine triangle normal and size
			const size_t v0=ID[neigsT[i]*3+0];
			const size_t v1=ID[neigsT[i]*3+1];
			const size_t v2=ID[neigsT[i]*3+2];

			//MT template for calculating normals: 2-1 x 0-2
			Mul(Pos[v2]-Pos[v1],Pos[v0]-Pos[v2],tmp);

			norm+=tmp.Ort();
		}
		w=float(neigsT.size());
	}

	//'w' should not be zero as any vertex should be part of some (at least one) triangle
	norm/=w;
}


void getVertexNormalviaTangents(const ActiveMesh& inMesh,
                                const size_t inV,
                                Vector3F& norm)
{
	//DOES NOT REALLY WORK, JUST TRIED AN IDEA I FOUND:
	//IT SEEMS TO (USUALLY) PROVIDE NORMALS SIMILAR TO WHAT WE CURRENTLY
	//GET WITH MT EXCEPT THAT SOME POINT INWARDS THE OBJECT

	//skip on incorrect input
	if (inV >= (signed)inMesh.Pos.GetLen()) return;

	//entry point to the list of vertices
	const Vector3FC* const Pos=inMesh.Pos.GetPtr();

	//list of surrounding neighbours to include in the calculation
	std::vector<size_t> neigs;
	getVertexImmediateNeighbours(inMesh,inV,neigs);

	//estimated tangent vectors
	Vector3F tu,tv;

	//helper variables...
	const float twoPIoverN=6.28f/float(neigs.size());
	const Vector3FC& v=Pos[inV];
	Vector3F tmp;

	for (size_t i=0; i < neigs.size(); ++i)
	{
		tmp=Pos[neigs[i]];
		tmp-=v;
		tu+=tmp * (float)cos(twoPIoverN * float(i));
		tv+=tmp * (float)sin(twoPIoverN * float(i));
	}

	Mul(tu,tv,tmp);
	norm=tmp.Ort();
}


//----------- vertex neighbours scanning functions -----------

void getVertexImmediateNeighbours(const ActiveMesh& inMesh,
                                  const size_t inV,
                                  std::vector<size_t>& neigs)
{
	//skip on incorrect input
	if (inV >= (signed)inMesh.Pos.GetLen()) return;

	//clear the output list(s)
	neigs.clear();

	//vital entry data: list of triangles and its length
	const DWORD* const ID=inMesh.ID.GetPtr();
	const size_t IDlen=inMesh.ID.GetLen();

	std::map<size_t, int> visited; //note: initiates with zero values

	for (size_t v=0; v < IDlen; ++v)
	if (ID[v] == inV)
	{
		//found a triangle inV is part of
		//
		//determine its neighbours
		size_t n1,n2=v%3;
		switch (n2) {
		case 0: n1=ID[v+1]; n2=ID[v+2]; break;
		case 1: n1=ID[v-1]; n2=ID[v+1]; break;
		case 2: n1=ID[v-2]; n2=ID[v-1]; break;
		}

		//add them on a neighbour list (if not already there)
		if (! visited[n1]) { neigs.push_back(n1); visited[n1]=1; }
		if (! visited[n2]) { neigs.push_back(n2); visited[n2]=1; }
	}
}


void getVertexNeighbours(const ActiveMesh& inMesh,
                         const size_t inV, const size_t maxLevel,
                         std::vector< std::vector<size_t> >& neigs)
{
	//skip on incorrect input
	if (inV >= (signed)inMesh.Pos.GetLen()) return;

	//clear the output list(s)
	neigs.clear();

	//vital entry data: list of triangles and its length
	const DWORD* const ID=inMesh.ID.GetPtr();
	const size_t IDlen=inMesh.ID.GetLen();

	//remembers all vertices that got discovered and at what level
	std::map<size_t, int> visited; //note: initiates with zero values

	//prepare zero-th and current level vertices list
	std::vector<size_t> neighbours;
	neighbours.reserve(20);
	neighbours.push_back(inV);
	visited[inV]=1;
	neigs.push_back(neighbours);

	//last processed level
	size_t lastLevel=0;

	while (lastLevel < neigs.size() && lastLevel < maxLevel)
	{
		//prepare a next-level neighbours list
		neighbours.clear();

		//iterate over all vertices from the previous level and check their neighbours
		for (size_t i=0; i < neigs[lastLevel].size(); ++i)
		{
			//sV = currently Searched Vertex
			const size_t sV=neigs[lastLevel][i];

			//find neighbours of the currently searched vertex
			for (size_t v=0; v < IDlen; ++v)
			if (ID[v] == sV)
			{
				//found a triangle sV is part of
				//
				//determine its neighbours
				size_t n1,n2=v%3;
				switch (n2) {
				case 0: n1=ID[v+1]; n2=ID[v+2]; break;
				case 1: n1=ID[v-1]; n2=ID[v+1]; break;
				case 2: n1=ID[v-2]; n2=ID[v-1]; break;
				}

				//add them on a neighbour list (if not already there)
				if (! visited[n1]) { neighbours.push_back(n1); visited[n1]=1; }
				if (! visited[n2]) { neighbours.push_back(n2); visited[n2]=1; }
			}
		}

		//add next level of neighbours (if there are still some)
		if (neighbours.size() > 0) neigs.push_back(neighbours);
		++lastLevel;
	}
}


void getVertexNeighboursUpToDist(const ActiveMesh& inMesh,
                                 const size_t inV,
                                 const float dist,
                                 std::vector<size_t>& neigs,
                                 std::vector<float>& dists)
{
	//TODO
}


void getVertexIncidentTriangles(const ActiveMesh& inMesh,
                                const size_t inV,
                                std::vector<size_t>& neigsT)
{
	//skip on incorrect input
	if (inV >= (signed)inMesh.Pos.GetLen()) return;

	//clear the output list(s)
	neigsT.clear();

	//vital entry data: list of triangles and its length
	const DWORD* const ID=inMesh.ID.GetPtr();
	const size_t IDlen=inMesh.ID.GetLen();

	std::map<size_t, int> visited; //note: initiates with zero values

	for (size_t v=0; v < IDlen; ++v)
	if (ID[v] == inV)
	{
		//found a triangle inV is part of
		const size_t tID=v/3;

		//add on a neighbour list (if not already there)
		if (! visited[tID]) { neigsT.push_back(tID); visited[tID]=1; }
	}
}


//----------- triangle neighbours scanning functions -----------

void getTriangleIncidentNeighbours(const ActiveMesh& inMesh,
                                   const size_t inT,
                                   std::vector<size_t>& neigsT,
                                   std::vector<float>& dists)
{
	//scans all triangles in the mesh and lists (in \e neigsT) those
	//that share at least one vertex with the input triangle;
	//\dist is in the same units as are the units of the vertex coordinates

	//vital entry data: lists of vertices and triangles and length
	const Vector3FC* const Pos=inMesh.Pos.GetPtr();
	const DWORD* const ID=inMesh.ID.GetPtr();
	const size_t IDlen=inMesh.ID.GetLen() /3;

	//skip on incorrect input
	if (inT >= (signed)IDlen) return;

	//reset the output list
	neigsT.clear();
	dists.clear();

	//this triangle's centre
	Vector3F centre;
	centre =Pos[ID[3*inT+0]];
	centre+=Pos[ID[3*inT+1]];
	centre+=Pos[ID[3*inT+2]];
	centre/=3.f;

	//this triangle's vertices offsets
	const size_t inV1=ID[3*inT+0];
	const size_t inV2=ID[3*inT+1];
	const size_t inV3=ID[3*inT+2];

	//over all triangles in the mesh:
	Vector3F tmp;
	for (size_t i=0; i < IDlen; ++i)
	{
		//skip over myself
		if (i == inT) continue;

		int hits=0;

		//min "offset distances" between vertices:
		//eval the first vertex...
		size_t d1=ID[3*i+0]-inV1;
		size_t d2=ID[3*i+0]-inV2;
		size_t d3=ID[3*i+0]-inV3;
		if (min( min(d1,d2),d3 ) == 0) ++hits;
		//wouldn't this be faster (fewer ifs)?
		//if (d1*d2*d3 == 0) ++hits;
		//may be perhaps even faster with bitwise-logic operators...

		//the second...
		d1=ID[3*i+1]-inV1;
		d2=ID[3*i+1]-inV2;
		d3=ID[3*i+1]-inV3;
		if (min( min(d1,d2),d3 ) == 0) ++hits;

		d1=ID[3*i+2]-inV1;
		d2=ID[3*i+2]-inV2;
		d3=ID[3*i+2]-inV3;
		if (min( min(d1,d2),d3 ) == 0) ++hits;

		//triangle fits into the definition of an immediate neighbour
		if (hits > 0)
		{
			//add this triangle's ID on the list
			neigsT.push_back(i);

			//calc triangle-centres distance and add on the list too
			tmp =Pos[ID[3*i+0]];
			tmp+=Pos[ID[3*i+1]];
			tmp+=Pos[ID[3*i+2]];
			tmp/=3.f;
			tmp-=centre;
			dists.push_back(tmp.Len());
		}
	}
}


void getTriangleNeighboursUpToDist(const ActiveMesh& inMesh,
                                   const size_t inT,
                                   const float dist,
                                   std::vector<size_t>& neigsT,
                                   std::vector<float>& dists)
{
	//scans all triangles in the mesh and lists (in \e neigsT) those
	//that have at least two vertices within the distance \e dist
	//from this (\e inT) triangle's centre; \dist is in the same
	//units as are the units of the vertex coordinates

	//vital entry data: lists of vertices and triangles and length
	const Vector3FC* const Pos=inMesh.Pos.GetPtr();
	const DWORD* const ID=inMesh.ID.GetPtr();
	const size_t IDlen=inMesh.ID.GetLen() /3;

	//skip on incorrect input
	if (inT >= (signed)IDlen) return;

	//reset the output list
	neigsT.clear();
	dists.clear();

	//this triangle's centre
	Vector3F centre;
	centre =Pos[ID[3*inT+0]];
	centre+=Pos[ID[3*inT+1]];
	centre+=Pos[ID[3*inT+2]];
	centre/=3.f;

	//time-saver :-)
	const float distSq=dist*dist;

	//over all triangles in the mesh:
	Vector3F tmp;
	for (size_t i=0; i < IDlen; ++i)
	{
		//skip over myself
		if (i == inT) continue;

		int hits=0;

		//eval the first vertex...
		tmp=Pos[ID[3*i+0]];
		tmp-=centre;
		if (tmp.LenQ() < distSq) hits++;

		//the second...
		tmp=Pos[ID[3*i+1]];
		tmp-=centre;
		if (tmp.LenQ() < distSq) hits++;

		tmp=Pos[ID[3*i+2]];
		tmp-=centre;
		if (tmp.LenQ() < distSq) hits++;

		//triangle fits into the definition of the neighbour
		if (hits >= 2)
		{
			//add this triangle's ID on the list
			neigsT.push_back(i);

			//calc triangle-centres distance and add on the list too
			tmp =Pos[ID[3*i+0]];
			tmp+=Pos[ID[3*i+1]];
			tmp+=Pos[ID[3*i+2]];
			tmp/=3.f;
			tmp-=centre;
			dists.push_back(tmp.Len());
		}
	}
}


//----------- triangle helper functions -----------

bool getTriangleWindiness(const ActiveMesh& inMesh,
                          const size_t inT)
{
	//skip on incorrect input
	if (inT >= (signed)inMesh.ID.GetLen()/3) return true;

	const Vector3FC& vPos0=inMesh.Pos[inMesh.ID[3*inT+0]];
	const Vector3FC& vPos1=inMesh.Pos[inMesh.ID[3*inT+1]];
	const Vector3FC& vPos2=inMesh.Pos[inMesh.ID[3*inT+2]];

	Vector3F r01=vPos1 - vPos0;
	Vector3F r12=vPos2 - vPos1;
	//TODO: must be compared against its stored normal
	//TODO: without normal or any other clue, the winding is impossible to determine

	return true;
}

} //namespace ulm
