#include "../cmath3d/TriangleMesh.h"

//rename MT MArray methods to std::vector their counterparts
#define GetPtr data
#define GetLen	size
#define DWORD long unsigned int

#ifndef __TRIANGLEMESH__3D_V__
#define __TRIANGLEMESH__3D_V__
#undef min
#undef max
#include <vector>
#define max(a, b) (a > b ? a : b)
#define min(a, b) (a < b ? a : b)
namespace ulm
{

//----------- vertex normal calculation functions -----------
/**
 * ala Dufour11 -- average of edge vectors (from the examined vertex outwards)
 * if \e maxLevel > 0, normals calculated for surrounding vertices contribute too
 *                     with weight decreasing with "distance" from the original vertex
 */
void getVertexNormalviaBarycentre(const ActiveMesh& inMesh,
                             const size_t inV,
                             Vector3F& norm,
                             const size_t maxLevel=1);


///Operational modes on how to combine multiple normal vectors in ulm::getVertexNormalviaTNormals()
#define NoWeight    0
#define AreaWeight  1
#define AngleWeight 2
/**
 Vertex normal is an average from normals of triangles incident to this vertex.
 To un-normalize the contributing normals, we weight them.
 If \e mode is NoWeight, all normals contribute equally strong.
 If \e mode is AreaWeight, triangle normals contribute linearly proportionally to the area of the triangle.
 If \e mode is AngleWeight, triangle normals contribute linearly proportionally according to the incident angle to the vertex.

 A variant of the \e mode=AngleWeight is basically implemented in TriangleMesh::_CalculateNode_Tension_UnNorm_Curv().
 In deed, the MT (MotionTracking) solution uses triangle weights by Nelson Max: Mean Weighted by Sine and Edge Length Reciprocal.

 The \e mode=AreaWeight is considered to be particularly popular/classical solution in computer graphics.
 However, the MT solution seems to provide more stable solution in the "cell imaging" world.
 */
void getVertexNormalviaTNormals(const ActiveMesh& inMesh,
                           const size_t inV,
                           Vector3F& norm,
                           const int mode=NoWeight);


/**
 vector product of two "tangent" vectors tu and tv, where
 tu = \sum_{i=0}^N p_i cos(2pi * i/N),
 tv = \sum_{i=0}^N p_i sin(2pi * i/N),
 p_i edge vectors to neighboring vertices

 DOES NOT REALLY WORK, JUST TRIED AN IDEA I FOUND:
 IT SEEMS TO (USUALLY) PROVIDE NORMALS SIMILAR TO WHAT WE CURRENTLY
 GET WITH MT EXCEPT THAT SOME NORMALS POINT INWARDS THE OBJECT
 */
void getVertexNormalviaTangents(const ActiveMesh& inMesh,
                           const size_t inV,
                           Vector3F& norm);


//----------- vertex neighbours scanning functions -----------
/**
 * Returns lists of neighbours of the given vertex \e inV of the mesh \e inMesh.
 * The discovered neighbours are stored in the list of \e neigs (in no particular specific order).
 * The \e neigs is cleared before the discovery starts.
 *
 * Note that vertices are represented as offsets in the vertex list of the mesh.
 */
void getVertexImmediateNeighbours(const ActiveMesh& inMesh,
                             const size_t inV,
                             std::vector<size_t>& neigs);


/**
 * Returns lists of neighbours of the given vertex \e inV of the mesh \e inMesh.
 * The neighbours are searched up to level \e maxLevel in a breadth-first search
 * manner, the immediate neighbours represent the level one (1) (are one hop away).
 * The discovered neighbours are stored in the "two-dimensional" \e neigs output
 * structure, e.g., \e neigs[2][3] contains 4th neighbour (counting from zero (0))
 * in the 2nd level. The \e neigs is cleared before the discovery starts.
 *
 * Note that vertices are represented as offsets in the vertex list of the mesh.
 * Note that \e neigs[0][0] == \e inV.
 * Note that \e neigs.size() <= \e maxLevel+1, '<' holds only when not so much levels were available.
 */
void getVertexNeighbours(const ActiveMesh& inMesh,
                    const size_t inV, const size_t maxLevel,
                    std::vector< std::vector<size_t> >& neigs);


/**
 * Lists all neighbours (vertices) that are reachable along edges from the input
 * vertex \e inV and are not farther as \e dist (in units of vertex position coordinates).
 * The function basically does the Dijkstra algorithms (variant of breadth-width
 * search).
 */
void  //TODO VLADO
getVertexNeighboursUpToDist(const ActiveMesh& inMesh,
                            const size_t inV,
						          const float dist,
                            std::vector<size_t>& neigs,
                            std::vector<float>& dists);


/**
 * Returns lists of triangle IDs the input vertex \e inV is part of.
 */
void getVertexIncidentTriangles(const ActiveMesh& inMesh,
                           const size_t inV,
                           std::vector<size_t>& neigsT);


//----------- triangle neighbours scanning functions -----------
/**
 * lists incident triangles in the mesh, i.e., those that
 * have at least one common vertex with input triangle \e inT
 *
 * output: \e neigsT is a list of neighboring triangle IDs;
 * \e dists is a corresponding list of centre-to-centre distances
 */
void getTriangleIncidentNeighbours(const ActiveMesh& inMesh,
                                   const size_t inT,
                                   std::vector<size_t>& neigsT,
                                   std::vector<float>& dists);


/**
 * scans all triangles in the mesh and lists (in \e neigs) those
 * that have at least two vertices within the distance \e dist
 * from this (\e inT) triangle's centre; \dist is in the same
 * units as are the units of the vertex coordinates
 *
 * output: \e neigsT is a list of triangle IDs that are within
 * the input \e dist distance; \e dists is a corresponding list
 * of centre-to-centre distances
 */
void getTriangleNeighboursUpToDist(const ActiveMesh& inMesh,
                                   const size_t inT,
                                   const float dist,
                                   std::vector<size_t>& neigsT,
                                   std::vector<float>& dists);


//----------- triangle helper functions -----------
/**
 * tells whether the order of triangle vertices is right-handed
 * (returns TRUE) or left-handed (returns FALSE). Right-handed
 * means that when moving in the direction from vertex 0. to vertex 1.,
 * we will turn right to move to vertex 2. from vertex 1.; in this case
 * the triangle has counter-clockwise order of vertex when observed
 * in the direction of triangle normal
 *
 * This is also called "triangle winding order".
 */
bool getTriangleWindiness(const ActiveMesh& inMesh,
                          const size_t inT);

} //namespace ulm

#endif
